
#ifndef __LISP_H__
#define __LISP_H__
#ifdef __cplusplus
extern "C" {
#endif


//---------------------------------------------------------------------------
#define SYMBOL_MAX_CNT				64
#define SYMBOL_MAX_BUF				16
#define CONS_MAX_CNT				4096
#define READ_MAX_BUF				128
#define PRINT_MAX_BUF				512

#define NIL 						((cell)0)
#define UNDEF						((cell)2)
#define SYMBOL_NAME(e)				(((ST_SYMBOL*)e)->name)
#define CAR(e)						(((ST_CONS*)e)->car)
#define CDR(e)						(((ST_CONS*)e)->cdr)
#define CAAR(e)						CAR(CAR(e))
#define CADR(e)						CAR(CDR(e))
#define CDAR(e)						CDR(CAR(e))
#define CDDR(e)						CDR(CDR(e))
#define CADDR(e)					CAR(CDR(CDR(e)))
#define CADDDR(e)					CAR(CDR(CDR(CDR(e))))

#define CELL2INT(e)					((s32)e >> 1)
#define INT2CELL(n)					((cell)(((s32)n << 1) + 1))

#define TOKEN_SPECIAL_STR			(u8*)"().'"

//---------------------------------------------------------------------------
typedef void* cell;


typedef struct {
	u8 name[SYMBOL_MAX_BUF];
} ST_SYMBOL;


typedef struct {
	cell car;
	cell cdr;
} ST_CONS;


typedef struct {
	u8* pFile;
	u8* pCur;
	u8* pCurPrev;
	u32 size;
	u16 cnt;
} ST_READ;


typedef struct {
	cell topEnv;
	// symbol
	cell symT;
	cell symFunc;
	cell symSform;
	// special-forms
	cell symQuote;
	cell symLambda;
	cell symMacro;
} ST_LISP;


typedef struct {
  u8  kind;
  u8* name;
  s32 num;
} ST_TOKEN;


//---------------------------------------------------------------------------
// initialize
EWRAM_CODE void LispInit();
EWRAM_CODE cell LispMakeSymbol(u8* name);
EWRAM_CODE cell LispFindSymbol(u8* name);
EWRAM_CODE void LispMakeSymbolVar(u8* name, cell* var);
EWRAM_CODE void LispMakeFunc(u8* name, void* func, s32 args);
EWRAM_CODE void LispMakeSform(u8* name, void* func);
EWRAM_CODE void LispMakeSformVar(u8* name, void* func, cell* var);
EWRAM_CODE cell LispMakeListCell(char* s);


// primitive functions
EWRAM_CODE cell LispFuncCar(cell e);
EWRAM_CODE cell LispFuncCdr(cell e);
EWRAM_CODE cell LispFuncCons(cell car, cell cdr);
EWRAM_CODE cell LispFuncConCat2(cell a, cell b);
EWRAM_CODE cell LispFuncAppend2(cell a, cell b);
EWRAM_CODE cell LispFuncAdd(cell a, cell b);
EWRAM_CODE cell LispFuncSub(cell a, cell b);
EWRAM_CODE cell LispFuncMul(cell a, cell b);
EWRAM_CODE cell LispFuncDiv(cell a, cell b);
EWRAM_CODE cell LispFuncEq(cell a, cell b);
EWRAM_CODE cell LispFuncGt(cell a, cell b);
EWRAM_CODE cell LispFuncAssoc(cell key, cell list);
EWRAM_CODE cell LispFuncAcons(cell key, cell value, cell alist);
EWRAM_CODE cell LispFuncPairlis(cell keys, cell values);
EWRAM_CODE cell LispFuncListF(cell list);
// special-form
EWRAM_CODE cell LispFuncQuote(cell e, cell env);
EWRAM_CODE cell LispFuncSetq(cell e, cell env);
EWRAM_CODE cell LispFuncCond(cell e, cell env);
EWRAM_CODE cell LispFuncLambda(cell e, cell env);
EWRAM_CODE cell LispFuncMacro(cell e, cell env);


// read
EWRAM_CODE ST_TOKEN* LispReadGetToken();
EWRAM_CODE u8   LispReadGetChr();
EWRAM_CODE void LispReadUnGetChr();
EWRAM_CODE cell LispReadNext(ST_TOKEN* token);
EWRAM_CODE cell LispReadCell();
EWRAM_CODE void LispReadFile(u8* pFile, u32 size);
EWRAM_CODE void LispReadSetBuf(u8* pBuf);
EWRAM_CODE void LispExec();



// print
EWRAM_CODE void LispSprintNext(u8* buf, cell e);
EWRAM_CODE u8*  LispSprintBuf(cell e);
EWRAM_CODE void LispPrintBuf(cell e);


// eval
EWRAM_CODE cell LispFuncEval(cell e, cell env);
EWRAM_CODE cell LispFuncEvalList(cell list, cell env);
EWRAM_CODE cell LispFuncApply(cell func, cell args, cell env);
EWRAM_CODE void LispChkArgCnt(cell f, s32 requires, s32 arg_count);


// macro
EWRAM_CODE bool LispIsInt(cell e);
EWRAM_CODE bool LispIsSymbol(cell e);
EWRAM_CODE bool LispIsCons(cell e);


// utility
EWRAM_CODE u16 LispGetlistLen(cell list);


// test tool
EWRAM_CODE void LispTest();
EWRAM_CODE void LispAssertStrEq(char* expected, char* value, char* msg);
EWRAM_CODE void LispAssertIntEq(s16 expected, s16 value, char* msg);
EWRAM_CODE void LispAssertChrEq(char* expected, u8 value, char* msg);
EWRAM_CODE void LispAssertLispEq(char* expected, cell value, char* msg);
EWRAM_CODE void LispAssertBool(bool is, char* msg);


EWRAM_CODE void LispTestMakeSymbol();
EWRAM_CODE void LispTestPrint();
EWRAM_CODE void LispTestToken();
EWRAM_CODE void LispTestRead();
EWRAM_CODE void LispTestFunction();
EWRAM_CODE void LispTestSpecialForm();
EWRAM_CODE void LispTestApplyEval();


#ifdef __cplusplus
}
#endif
#endif
